@@ -28,6 +28,10 @@
 #include <stdint.h>
 #include <unistd.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#if defined(__GNU_LIBRARY__)
+#include <sys/sysmacros.h>		// for major, minor
+#endif
 #include <ftw.h>
 #include <grp.h>
 #include <pwd.h>
@@ -37,6 +41,9 @@
 #include "lzip.h"
 #include "tarlz.h"
 
+
+const CRC32C crc32c;
+
 int cl_owner = -1;		// global vars needed by add_member
 int cl_group = -1;
 int cl_solid = 0;		// 1 = dsolid, 2 = asolid, 3 = solid
@@ -133,11 +140,89 @@ bool archive_write( const uint8_t * const buf, const int size )
   }
 
 
+unsigned char xdigit( const unsigned value )
+  {
+  if( value <= 9 ) return '0' + value;
+  if( value <= 15 ) return 'A' + value - 10;
+  return 0;
+  }
+
+void print_hex( char * const buf, int size, unsigned long long num )
+  {
+  while( --size >= 0 ) { buf[size] = xdigit( num & 0x0F ); num >>= 4; }
+  }
+
 void print_octal( char * const buf, int size, unsigned long long num )
   {
   while( --size >= 0 ) { buf[size] = '0' + ( num % 8 ); num /= 8; }
   }
 
+unsigned decimal_digits( unsigned long long value )
+  {
+  unsigned digits = 1;
+  while( value >= 10 ) { value /= 10; ++digits; }
+  return digits;
+  }
+
+unsigned long long record_size( const unsigned keyword_size,
+                                const unsigned long long value_size )
+  {
+  // size = ' ' + keyword + '=' + value + '\n'
+  const unsigned long long size = 1 + keyword_size + 1 + value_size + 1;
+  const unsigned d1 = decimal_digits( size );
+  return decimal_digits( d1 + size ) + size;
+  }
+
+bool write_extended( const Extended & extended, const Tar_header uheader )
+  {
+  const int path_rec = extended.path.size() ?
+                       record_size( 4, extended.path.size() ) : 0;
+  const int lpath_rec = extended.linkpath.size() ?
+                        record_size( 8, extended.linkpath.size() ) : 0;
+  const int size_rec = ( extended.size > 0 ) ?
+                       record_size( 4, decimal_digits( extended.size ) ) : 0;
+  const unsigned long long edsize = path_rec + lpath_rec + size_rec + 22;
+  const unsigned long long bufsize = round_up( edsize );
+  if( bufsize == 0 ) return edsize == 0;	// overflow or no extended data
+  char * const buf = new char[bufsize+1];	// extended records buffer
+  unsigned long long pos = path_rec;		// goto can't cross this
+  if( path_rec && snprintf( buf, path_rec + 1, "%d path=%s\n",
+                            path_rec, extended.path.c_str() ) != path_rec )
+    goto error;
+  if( lpath_rec && snprintf( buf + pos, lpath_rec + 1, "%d linkpath=%s\n",
+                     lpath_rec, extended.linkpath.c_str() ) != lpath_rec )
+    goto error;
+  pos += lpath_rec;
+  if( size_rec && snprintf( buf + pos, size_rec + 1, "%d size=%llu\n",
+                            size_rec, extended.size ) != size_rec )
+    goto error;
+  pos += size_rec;
+  if( snprintf( buf + pos, 23, "22 GNU.crc32=00000000\n" ) != 22 ) goto error;
+  pos += 22;
+  if( pos != edsize ) goto error;
+  print_hex( buf + edsize - 9, 8,
+             crc32c.windowed_crc( (const uint8_t *)buf, edsize - 9, edsize ) );
+  std::memset( buf + edsize, 0, bufsize - edsize );	// wipe padding
+  Tar_header header;				// extended header
+  std::memcpy( header, uheader, header_size );	// copy from ustar header
+  header[typeflag_o] = tf_extended;		// then modify some fields
+  print_octal( header + size_o, size_l - 1, edsize );
+  print_octal( header + chksum_o, chksum_l - 1,
+               ustar_chksum( (const uint8_t *)header ) );
+  if( !archive_write( (const uint8_t *)header, header_size ) ) goto error;
+  for( pos = 0; pos < bufsize; )	// write extended records to archive
+    {
+    int size = std::min( bufsize - pos, 1ULL << 20 );
+    if( !archive_write( (const uint8_t *)buf + pos, size ) ) goto error;
+    pos += size;
+    }
+  delete[] buf;
+  return true;
+error:
+  delete[] buf;
+  return false;
+  }
+
 
 const char * remove_leading_dotdot( const char * const filename )
   {
@@ -164,7 +249,8 @@ const char * remove_leading_dotdot( const char * const filename )
   }
 
 
-bool split_name( const char * const filename, Tar_header header )
+bool store_name( const char * const filename, Extended & extended,
+                 Tar_header header )
   {
   const char * const stored_name = remove_leading_dotdot( filename );
   const int len = std::strlen( stored_name );
@@ -173,15 +259,19 @@ bool split_name( const char * const filename, Tar_header header )
     { std::memcpy( header + name_o, stored_name, len ); return true; }
   if( len <= max_len )				// find shortest prefix
     for( int i = len - name_l - 1; i < len && i <= prefix_l; ++i )
-      if( stored_name[i] == '/' )
+      if( stored_name[i] == '/' )		// stored_name can be split
         {
         std::memcpy( header + name_o, stored_name + i + 1, len - i - 1 );
         std::memcpy( header + prefix_o, stored_name, i );
         return true;
         }
+  // store full name in extended record, truncated name in ustar header
+  extended.path = stored_name;
+  std::strncpy( header + name_o, stored_name, name_l );
   return false;
   }
 
+
 int add_member( const char * const filename, const struct stat *,
                 const int flag, struct FTW * )
   {
@@ -189,11 +279,10 @@ int add_member( const char * const filename, const struct stat *,
   if( lstat( filename, &st ) != 0 )
     { show_file_error( filename, "Can't stat input file", errno );
       gretval = 1; return 0; }
+  Extended extended;		// metadata for extended records
   Tar_header header;
   std::memset( header, 0, header_size );
-  if( !split_name( filename, header ) )
-    { show_file_error( filename, "File name is too long." );
-      gretval = 2; return 0; }
+  store_name( filename, extended, header );
 
   const mode_t mode = st.st_mode;
   print_octal( header + mode_o, mode_l - 1,
@@ -220,8 +309,13 @@ int add_member( const char * const filename, const struct stat *,
     if( st.st_size > linkname_l ||
         readlink( filename, header + linkname_o, linkname_l ) != st.st_size )
       {
-      show_file_error( filename, "Link destination name is too long." );
-      gretval = 2; return 0;
+      char * const buf = new char[st.st_size+1];
+      const long len = readlink( filename, buf, st.st_size );
+      if( len == st.st_size ) { buf[len] = 0; extended.linkpath = buf; }
+      delete[] buf;
+      if( len != st.st_size )
+      { show_file_error( filename, "Error reading link", (len < 0) ? errno : 0 );
+        gretval = 2; return 0; }
       }
     }
   else if( S_ISCHR( mode ) || S_ISBLK( mode ) )
@@ -242,14 +336,17 @@ int add_member( const char * const filename, const struct stat *,
   const struct group * const gr = getgrgid( gid );
   if( gr && gr->gr_name )
     std::strncpy( header + gname_o, gr->gr_name, gname_l - 1 );
-  print_octal( header + size_o, size_l - 1, file_size );
+  if( file_size >= 1ULL << 33 ) extended.size = file_size;
+  else print_octal( header + size_o, size_l - 1, file_size );
   print_octal( header + chksum_o, chksum_l - 1,
                ustar_chksum( (const uint8_t *)header ) );
 
   const int infd = file_size ? open_instream( filename ) : -1;
   if( file_size && infd < 0 ) { gretval = 1; return 0; }
+  if( !extended.empty() && !write_extended( extended, header ) )
+    { show_error( "Error writing extended header", errno ); return 1; }
   if( !archive_write( (const uint8_t *)header, header_size ) )
-    { show_error( "Error writing archive header", errno ); return 1; }
+    { show_error( "Error writing ustar header", errno ); return 1; }
   if( file_size )
     {
     enum { bufsize = 32 * header_size };
