@@ -20,6 +20,7 @@
 #include <algorithm>
 #include <cerrno>
 #include <cstdio>
+#include <cstdlib>
 #include <cstring>
 #include <string>
 #include <vector>
@@ -27,6 +28,10 @@
 #include <unistd.h>
 #include <utime.h>
 #include <sys/stat.h>
+#include <sys/types.h>
+#if defined(__GNU_LIBRARY__)
+#include <sys/sysmacros.h>		// for makedev
+#endif
 #include <lzlib.h>
 
 #include "arg_parser.h"
@@ -250,50 +255,38 @@ const char * user_group_string( const Tar_header header )
   }
 
 
-const char * link_string( const Tar_header header )
-  {
-  enum { bufsize = 9 + linkname_l + 1 };
-  static char buf[bufsize];
-  const Typeflag typeflag = (Typeflag)header[typeflag_o];
-
-  if( typeflag == tf_link )
-    snprintf( buf, bufsize, " link to %.100s", header + linkname_o );
-  else if( typeflag == tf_symlink )
-    snprintf( buf, bufsize, " -> %.100s", header + linkname_o );
-  else buf[0] = 0;
-  return buf;
-  }
-
-
-void show_member_name( const char * const filename, const Tar_header header,
+void show_member_name( const Extended & extended, const Tar_header header,
                        const int vlevel )
   {
   if( verbosity < vlevel ) return;
   if( verbosity > vlevel )
     {
     const time_t mtime = strtoull( header + mtime_o, 0, 8 );	// 33 bits
-    struct tm * tm = localtime( &mtime );
-    std::printf( "%s %s %9llu %4d-%02u-%02u %02u:%02u %s%s\n",
+    const struct tm * const tm = localtime( &mtime );
+    const Typeflag typeflag = (Typeflag)header[typeflag_o];
+    const bool islink = ( typeflag == tf_link || typeflag == tf_symlink );
+    const char * const link_string = !islink ? "" :
+                         ( ( typeflag == tf_link ) ? " link to " : " -> " );
+    std::printf( "%s %s %9llu %4d-%02u-%02u %02u:%02u %s%s%s\n",
                  mode_string( header ), user_group_string( header ),
-                 strtoull( header + size_o, 0, 8 ), 1900 + tm->tm_year,
-                 1 + tm->tm_mon, tm->tm_mday, tm->tm_hour, tm->tm_min,
-                 filename, link_string( header ) );
+                 extended.size, 1900 + tm->tm_year, 1 + tm->tm_mon,
+                 tm->tm_mday, tm->tm_hour, tm->tm_min, extended.path.c_str(),
+                 link_string, !islink ? "" : extended.linkpath.c_str() );
     }
-  else std::printf( "%s\n", filename );
+  else std::printf( "%s\n", extended.path.c_str() );
   std::fflush( stdout );
   }
 
 
-int list_member( const int infd, const char * const filename,
-                 const unsigned long long file_size, const Tar_header header,
-                 const bool skip )
+int list_member( const int infd, const Extended & extended,
+                 const Tar_header header, const bool skip )
   {
-  if( !skip ) show_member_name( filename, header, 0 );
+  if( !skip ) show_member_name( extended, header, 0 );
 
   const unsigned bufsize = 32 * header_size;
   uint8_t buf[bufsize];
-  unsigned long long rest = file_size;
-  const int rem = file_size % header_size;
+  unsigned long long rest = extended.size;
+  const int rem = extended.size % header_size;
   const int padding = rem ? header_size - rem : 0;
   while( rest > 0 )
     {
@@ -316,13 +309,14 @@ bool contains_dotdot( const char * const filename )
   }
 
 
-int extract_member( const int infd, const char * const filename,
-                    const unsigned long long file_size, const Tar_header header )
+int extract_member( const int infd, const Extended & extended,
+                    const Tar_header header )
   {
+  const char * const filename = extended.path.c_str();
   if( contains_dotdot( filename ) )
     {
     show_file_error( filename, "Contains a '..' component, skipping." );
-    return list_member( infd, filename, file_size, header, true );
+    return list_member( infd, extended, header, true );
     }
   const mode_t mode = strtoul( header + mode_o, 0, 8 );		// 12 bits
   const time_t mtime = strtoull( header + mtime_o, 0, 8 );	// 33 bits
@@ -330,7 +324,7 @@ int extract_member( const int infd, const char * const filename,
   const bool islink = ( typeflag == tf_link || typeflag == tf_symlink );
   int outfd = -1;
 
-  show_member_name( filename, header, 1 );
+  show_member_name( extended, header, 1 );
   std::remove( filename );
   make_path( filename );
   switch( typeflag )
@@ -344,14 +338,12 @@ int extract_member( const int infd, const char * const filename,
     case tf_link:
     case tf_symlink:
       {
-      char linkname[linkname_l+1];
-      std::memcpy( linkname, header + linkname_o, linkname_l );
-      linkname[linkname_l] = 0;
+      const char * const linkname = extended.linkpath.c_str();
 /*      if( contains_dotdot( linkname ) )
         {
         show_file_error( filename,
           "Link destination contains a '..' component, skipping." );
-        return list_member( infd, filename, file_size, header, false );
+        return list_member( infd, extended, header, false );
         }*/
       const bool hard = typeflag == tf_link;
       if( ( hard && link( linkname, filename ) != 0 ) ||
@@ -400,7 +392,8 @@ int extract_member( const int infd, const char * const filename,
 
   const uid_t uid = (uid_t)strtoul( header + uid_o, 0, 8 );
   const gid_t gid = (gid_t)strtoul( header + gid_o, 0, 8 );
-  if( !islink && chown( filename, uid, gid ) != 0 && errno != EPERM )
+  if( !islink && chown( filename, uid, gid ) != 0 &&
+      errno != EPERM && errno != EINVAL )
     {
     show_file_error( filename, "Can't change file owner", errno );
     return 2;
@@ -408,8 +401,8 @@ int extract_member( const int infd, const char * const filename,
 
   const unsigned bufsize = 32 * header_size;
   uint8_t buf[bufsize];
-  unsigned long long rest = file_size;
-  const int rem = file_size % header_size;
+  unsigned long long rest = extended.size;
+  const int rem = extended.size % header_size;
   const int padding = rem ? header_size - rem : 0;
   while( rest > 0 )
     {
@@ -477,28 +470,91 @@ bool compare_tslash( const char * const name1, const char * const name2 )
 } // end namespace
 
 
+bool Extended::parse( const int infd, const Tar_header header,
+                      const bool permissive )
+  {
+  const unsigned long long edsize = strtoull( header + size_o, 0, 8 );
+  const unsigned long long bufsize = round_up( edsize );
+  if( bufsize == 0 || edsize == 0 || edsize >= 1ULL << 33 )
+    return false;				// overflow or no extended data
+  char * const buf = new char[bufsize];		// extended records buffer
+  if( !archive_read( infd, (uint8_t *)buf, bufsize ) ) goto error;
+  for( unsigned long long pos = 0; pos < edsize; )	// parse records
+    {
+    char * tail;
+    const unsigned long long rsize = strtoull( buf + pos, &tail, 10 );
+    if( rsize == 0 || rsize > edsize - pos || tail[0] != ' ' ||
+        buf[pos+rsize-1] != '\n' ) goto error;
+    ++tail;	// point to keyword
+    // length of (keyword + '=' + value) without the final newline
+    const unsigned long long rest = ( buf + pos + rsize - 1 ) - tail;
+    if( rest > 5 && std::memcmp( tail, "path=", 5 ) == 0 )
+      { if( path.size() && !permissive ) goto error;
+        path.assign( tail + 5, rest - 5 ); }
+    else if( rest > 9 && std::memcmp( tail, "linkpath=", 9 ) == 0 )
+      { if( linkpath.size() && !permissive ) goto error;
+        linkpath.assign( tail + 9, rest - 9 ); }
+    else if( rest > 5 && std::memcmp( tail, "size=", 5 ) == 0 )
+      {
+      if( size != 0 && !permissive ) goto error;
+      size = 0;
+      for( unsigned long long i = 5; i < rest; ++i )
+        {
+        if( tail[i] < '0' || tail[i] > '9' ) goto error;
+        const unsigned long long prev = size;
+        size = size * 10 + ( tail[i] - '0' );
+        if( size < prev ) goto error;		// overflow
+        }
+      if( size < 1ULL << 33 ) goto error;	// size fits in ustar header
+      }
+    else if( rest > 10 && std::memcmp( tail, "GNU.crc32=", 10 ) == 0 )
+      {
+      if( crc_present && !permissive ) goto error;
+      if( rsize != 22 ) goto error;
+      char * t;
+      const uint32_t stored_crc = strtoul( tail + 10, &t, 16 );
+      if( t - tail - 10 != 8 || t[0] != '\n' ) goto error;
+      const uint32_t computed_crc =
+        crc32c.windowed_crc( (const uint8_t *)buf, pos + rsize - 9, edsize );
+      crc_present = true;
+      if( stored_crc != computed_crc ) goto error;
+      }
+    pos += rsize;
+    }
+  delete[] buf;
+  return true;
+error:
+  delete[] buf;
+  return false;
+  }
+
+
 int decode( const std::string & archive_name, const Arg_parser & parser,
-            const int filenames, const bool listing )
+            const int filenames, const bool ignore_crc, const bool listing,
+            const bool permissive )
   {
   const int infd = archive_name.size() ?
                    open_instream( archive_name ) : STDIN_FILENO;
   if( infd < 0 ) return 1;
 
+  // execute -C options and mark filenames to be extracted or listed
   std::vector< bool > name_pending( parser.arguments(), false );
   for( int i = 0; i < parser.arguments(); ++i )
     {
     const int code = parser.code( i );
     if( code == 'C' && !listing )
       {
-      const char * const filename = parser.argument( i ).c_str();
-      if( chdir( filename ) != 0 )
-        { show_file_error( filename, "Error changing working directory", errno );
+      const char * const dir = parser.argument( i ).c_str();
+      if( chdir( dir ) != 0 )
+        { show_file_error( dir, "Error changing working directory", errno );
           return 1; }
       }
     if( !code ) name_pending[i] = true;
     }
 
+  Extended extended;		// metadata from extended records
   int retval = 0;
+  bool prev_extended = false;	// prev header was extended
   bool skipping = false;
   while( true )			// process one member per iteration
     {
@@ -515,17 +571,46 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
     skipping = false;
 
     const char * const header = (const char *)buf;
-    enum { max_filename_size = prefix_l + 1 + name_l + 1 };
-    char stored_name[max_filename_size];
-    int len = 0;
-    while( len < prefix_l && header[prefix_o+len] )
-      { stored_name[len] = header[prefix_o+len]; ++len; }
-    if( len && header[name_o] ) stored_name[len++] = '/';
-    for( int i = 0; i < name_l && header[name_o+i]; ++i )
-      { stored_name[len] = header[name_o+i]; ++len; }
-    while( len > 0 && stored_name[len-1] == '/' ) --len;	// trailing '/'
-    stored_name[len] = 0;
-    const char * const filename = remove_leading_slash( stored_name );
+    const Typeflag typeflag = (Typeflag)header[typeflag_o];
+    if( typeflag == tf_extended )
+      {
+      if( prev_extended && !permissive )
+        { show_error( "Format violation: consecutive extended headers found."
+                      /*" Use --permissive."*/, 0, true ); return 2; }
+      if( !extended.parse( infd, header, permissive ) )
+        { show_error( "Error in extended records. Skipping to next header." );
+          extended.reset(); gretval = 2; }
+      else if( !extended.crc_present && !ignore_crc )
+        { show_error( "Missing CRC in extended records. Use --ignore-crc.",
+                      0, true ); return 2; }
+      prev_extended = true;
+      continue;
+      }
+    prev_extended = false;
+
+    if( extended.linkpath.empty() )
+      {
+      for( int i = 0; i < linkname_l && header[linkname_o+i]; ++i )
+        extended.linkpath += header[linkname_o+i];
+      while( extended.linkpath.size() > 1 &&		// trailing '/'
+             extended.linkpath[extended.linkpath.size()-1] == '/' )
+        extended.linkpath.resize( extended.linkpath.size() - 1 );
+      }
+
+    if( extended.path.empty() )
+      {
+      char stored_name[prefix_l+1+name_l+1];
+      int len = 0;
+      while( len < prefix_l && header[prefix_o+len] )
+        { stored_name[len] = header[prefix_o+len]; ++len; }
+      if( len && header[name_o] ) stored_name[len++] = '/';
+      for( int i = 0; i < name_l && header[name_o+i]; ++i )
+        { stored_name[len] = header[name_o+i]; ++len; }
+      while( len > 0 && stored_name[len-1] == '/' ) --len;	// trailing '/'
+      stored_name[len] = 0;
+      extended.path = remove_leading_slash( stored_name );
+      }
+    const char * const filename = extended.path.c_str();
 
     bool skip = filenames > 0;
     if( skip )
@@ -535,14 +620,15 @@ int decode( const std::string & archive_name, const Arg_parser & parser,
               compare_tslash( filename, parser.argument( i ).c_str() ) ) )
           { skip = false; name_pending[i] = false; break; }
 
-    const Typeflag typeflag = (Typeflag)header[typeflag_o];
-    const unsigned long long file_size =
-      ( typeflag == tf_regular || typeflag == tf_hiperf ) ?
-        strtoull( header + size_o, 0, 8 ) : 0;
+    if( extended.size == 0 &&
+        ( typeflag == tf_regular || typeflag == tf_hiperf ) )
+      extended.size = strtoull( header + size_o, 0, 8 );
+
     if( listing || skip )
-      retval = list_member( infd, filename, file_size, header, skip );
+      retval = list_member( infd, extended, header, skip );
     else
-      retval = extract_member( infd, filename, file_size, header );
+      retval = extract_member( infd, extended, header );
+    extended.reset();
     if( retval ) return retval;
     }
 
