@@ -34,12 +34,20 @@
 #include <fcntl.h>
 #include <stdint.h>
 #include <unistd.h>
+#include <sys/stat.h>
 #include <grp.h>
 #include <pwd.h>
+#if defined(__OS2__)
+#include <io.h>
+#endif
 
 #include "arg_parser.h"
 #include "tarlz.h"
 
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
 #if CHAR_BIT != 8
 #error "Environments where CHAR_BIT != 8 are not supported."
 #endif
@@ -77,13 +85,15 @@ void show_help()
                "      --solid                create solidly compressed archive\n"
                "      --group=<group>        use <group> name/id for added files\n"
                "      --owner=<owner>        use <owner> name/id for added files\n"
+               "      --ignore-crc           ignore missing CRC in extended records\n"
+//               "      --permissive           allow repeated extended headers and records\n"
                "      --uncompressed         don't compress the created archive\n"
                "\nExit status: 0 for a normal exit, 1 for environmental problems (file\n"
                "not found, invalid flags, I/O errors, etc), 2 to indicate a corrupt or\n"
                "invalid input file, 3 for an internal consistency error (eg, bug) which\n"
                "caused tarlz to panic.\n"
                "\nReport bugs to lzip-bug@nongnu.org\n"
-               "Lzip home page: http://www.nongnu.org/lzip/lzip.html\n" );
+               "Tarlz home page: http://www.nongnu.org/lzip/tarlz.html\n" );
   }
 
 
@@ -179,7 +189,7 @@ void set_group( const char * const arg )
 
 int open_instream( const std::string & name )
   {
-  const int infd = open( name.c_str(), O_RDONLY );
+  const int infd = open( name.c_str(), O_RDONLY | O_BINARY );
   if( infd < 0 )
     show_file_error( name.c_str(), "Can't open for reading", errno );
   return infd;
@@ -188,7 +198,7 @@ int open_instream( const std::string & name )
 
 int open_outstream( const std::string & name, const bool create )
   {
-  const int flags = create ? O_CREAT | O_WRONLY | O_TRUNC : O_RDWR;
+  const int flags = (create ? O_CREAT | O_WRONLY | O_TRUNC : O_RDWR) | O_BINARY;
   const mode_t outfd_mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;
 
   const int outfd = open( name.c_str(), flags, outfd_mode );
@@ -252,10 +262,10 @@ void show_error( const char * const msg, const int errcode, const bool help )
 void show_file_error( const char * const filename, const char * const msg,
                       const int errcode )
   {
-  if( verbosity < 0 ) return;
-  std::fprintf( stderr, "%s: %s: %s", program_name, filename, msg );
-  if( errcode > 0 ) std::fprintf( stderr, ": %s", std::strerror( errcode ) );
-  std::fputc( '\n', stderr );
+  if( verbosity >= 0 )
+    std::fprintf( stderr, "%s: %s: %s%s%s\n", program_name, filename, msg,
+                  ( errcode > 0 ) ? ": " : "",
+                  ( errcode > 0 ) ? std::strerror( errcode ) : "" );
   }
 
 
@@ -272,9 +282,12 @@ int main( const int argc, const char * const argv[] )
   std::string archive_name;
   Mode program_mode = m_none;
   int level = 6;		// compression level, < 0 = uncompressed
+  bool ignore_crc = false;
+  bool permissive = false;
   invocation_name = argv[0];
 
-  enum { opt_aso = 256, opt_dso, opt_grp, opt_own, opt_sol, opt_un };
+  enum { opt_aso = 256, opt_crc, opt_dso, opt_grp, opt_own, opt_per, opt_sol,
+         opt_un };
   const Arg_parser::Option options[] =
     {
     { '0',  0,                Arg_parser::no  },
@@ -301,7 +314,9 @@ int main( const int argc, const char * const argv[] )
     { opt_aso, "asolid",      Arg_parser::no  },
     { opt_dso, "dsolid",      Arg_parser::no  },
     { opt_grp, "group",       Arg_parser::yes },
+    { opt_crc, "ignore-crc",  Arg_parser::no  },
     { opt_own, "owner",       Arg_parser::yes },
+    { opt_per, "permissive",  Arg_parser::no  },
     { opt_sol, "solid",       Arg_parser::no  },
     { opt_un, "uncompressed", Arg_parser::no  },
     {  0 ,  0,                Arg_parser::no  } };
@@ -334,23 +349,30 @@ int main( const int argc, const char * const argv[] )
       case 'V': show_version(); return 0;
       case 'x': set_mode( program_mode, m_extract ); break;
       case opt_aso: cl_solid = 2; break;
+      case opt_crc: ignore_crc = true; break;
       case opt_dso: cl_solid = 1; break;
       case opt_grp: set_group( arg ); break;
       case opt_own: set_owner( arg ); break;
+      case opt_per: permissive = true; break;
       case opt_sol: cl_solid = 3; break;
       case opt_un: level = -1; break;
       default : internal_error( "uncaught option" );
       }
     } // end process options
 
+#if defined(__MSVCRT__) || defined(__OS2__)
+  setmode( STDIN_FILENO, O_BINARY );
+  setmode( STDOUT_FILENO, O_BINARY );
+#endif
+
   switch( program_mode )
     {
-    case m_none: show_error( "Missing operation.", 0, true ); return 2;
+    case m_none:   show_error( "Missing operation.", 0, true ); return 2;
     case m_append:
     case m_create: return encode( archive_name, parser, filenames, level,
                                   program_mode == m_append );
     case m_extract:
-    case m_list:
-      return decode( archive_name, parser, filenames, program_mode == m_list );
+    case m_list:   return decode( archive_name, parser, filenames, ignore_crc,
+                                  program_mode == m_list, permissive );
     }
   }
